{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE FlexibleContexts #-}
module Text.TEIWA.Source (
      Format
    , Origin(..)
    , TEIWAParser
    , Source(..)
    , coNLLX
    , csv
    , runTEIWAParser
    , parse
    , tsv
  ) where

import Control.Monad.Except (MonadError(..))
import Control.Monad.IO.Class (MonadIO(..))
import Data.Text.Lazy as Text (Text, unpack)
import Data.Text.Lazy.IO as Text (readFile)
import Text.Parsec (ParsecT, SourceName, runParserT)
import Text.TEIWA.Annotation (
    Annotation(..), SentenceAnnotation(..), TokenAnnotation(..)
  )
import Text.TEIWA.Config (Config(..))
import Text.TEIWA.Error (Error(..))
import Text.TEIWA.Source.Common (AnnotationContext(..), Row)
import qualified Text.TEIWA.Source.ConLLX as ConLLX (getContext, sentences)
import qualified Text.TEIWA.Source.CSV as CSV (body, getContext)

type TEIWAParser = ParsecT Text () (Either Error)
type Format = Config -> TEIWAParser Annotation

annotateToken :: MonadError Error m =>
  AnnotationContext -> Row -> m TokenAnnotation
annotateToken (AnnotationContext {columnIndex, columnName, header}) (atLine, record) =
  case splitAt columnIndex record of
    (_, []) -> throwError . MissingColumn atLine $ Text.unpack columnName
    (before, form:after) ->
      return $ TokenAnnotation {form, annotated = zip header (before ++ after)}

coNLLX :: Format
coNLLX (Config {formColumn}) = do
  context <- ConLLX.getContext formColumn
  SentenceLevel <$> (
      ConLLX.sentences >>= mapM (
          fmap SentenceAnnotation . mapM (annotateToken context)
        )
    )

csv :: Format
csv (Config {formColumn}) = do
  context <- CSV.getContext ',' formColumn
  TokenLevel <$> (CSV.body ',' >>= mapM (annotateToken context))

tsv :: Format
tsv (Config {formColumn}) = do
  context <- CSV.getContext '\t' formColumn
  TokenLevel <$> (CSV.body '\t' >>= mapM (annotateToken context))

data Origin = File FilePath | Text Text

data Source = Source {
      format :: Format
    , origin :: Origin
  }

runTEIWAParser :: MonadError Error m =>
  TEIWAParser a -> SourceName -> Text -> m a
runTEIWAParser p s = flattenErrors . runParserT p () s
  where
    flattenErrors = either throwError (either (throwError . ParsingError) pure)

parse :: (MonadIO m, MonadError Error m) =>
  Config -> Source -> m Annotation
parse config (Source {format, origin}) = parseFrom (format config) origin
  where
    parseFrom p (File f) = liftIO (Text.readFile f) >>= runTEIWAParser p f
    parseFrom p (Text t) = runTEIWAParser p "" t
